/* -*-  Mode:C++; c-basic-offset:4; tab-width:4; indent-tabs-mode:nil -*- */
/*
 * Copyright (C) 2016-2024 by the University of Southern California
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <cassert>
#include <list>
#include <string>
#include <functional>
#include <libtrace.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <stdbool.h>
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <grp.h>
#include <pwd.h>

#include "mtc_log.hh"
#include "mtc_output.hh"

#define MAXWAIT_MS 1.0

using namespace std;

// special type for controlling snaplen when it depends on the packet headers
typedef int slen_t;
// special values of snaplen for snaplen_filter_t
const slen_t SNAPLEN_UPTO_IP  = slen_t(-1);
const slen_t SNAPLEN_UPTO_IP6 = slen_t(-2);
const slen_t SNAPLEN_UPTO_TCP = slen_t(-3);
const slen_t SNAPLEN_UPTO_UDP = slen_t(-4);

const size_t SLEN_ERROR = size_t(-1);

typedef function<slen_t (const libtrace_packet_t *)> slenfunc_t;
struct snaplen_filter_t {
    struct libtrace_filter_t *filter;
    slenfunc_t               sfunc;
    snaplen_filter_t(struct libtrace_filter_t *f, slenfunc_t fp): filter(f), sfunc(fp){}
};

static slenfunc_t slen_const(size_t slen) {
    return [slen](const libtrace_packet_t *p) {
        return slen;
    };
}

static slenfunc_t slen_name2func(string name, size_t extra) {
    if (name == "UPTO_IP") {
        return [extra](const libtrace_packet_t *p) {
            libtrace_ip_t *ip = trace_get_ip(const_cast<libtrace_packet_t*>(p));
            uint32_t slen, rem;
            char *first_byte = (char*)trace_get_packet_buffer(p, NULL, &slen);
            if (ip != NULL) {
                rem = slen - ((char*)ip - first_byte);
                trace_get_payload_from_ip(ip, NULL, &rem);
                return size_t(slen-rem) + extra;
            }
            return SLEN_ERROR;
        };
    } else if (name == "UPTO_IP6") {
        return [extra](const libtrace_packet_t *p) {
            libtrace_ip6_t *ip6 = trace_get_ip6(const_cast<libtrace_packet_t*>(p));
            uint32_t slen, rem;
            char *first_byte = (char*)trace_get_packet_buffer(p, NULL, &slen);
            if (ip6 != NULL) {
                rem = slen - ((char*)ip6 - first_byte);
                trace_get_payload_from_ip6(ip6, NULL, &rem);
                return size_t(slen-rem) + extra;
            }
            return SLEN_ERROR;
        };
    } else if (name == "UPTO_TCP") {
        return [extra](const libtrace_packet_t *p) {
            libtrace_tcp_t *tcp = trace_get_tcp(const_cast<libtrace_packet_t*>(p));
            uint32_t slen, rem;
            char *first_byte = (char*)trace_get_packet_buffer(p, NULL, &slen);
            if (tcp != NULL) {
                rem = slen - ((char*)tcp - first_byte);
                trace_get_payload_from_tcp(tcp, &rem);
                return size_t(slen-rem)+extra;
            }
            return SLEN_ERROR;
        };
    } else if (name == "UPTO_UDP") {
        return [extra](const libtrace_packet_t *p) {
            libtrace_udp_t *udp = trace_get_udp(const_cast<libtrace_packet_t*>(p));
            uint32_t slen, rem;
            char *first_byte = (char*)trace_get_packet_buffer(p, NULL, &slen);
            if (udp != NULL) {
                rem = slen - ((char*)udp - first_byte);
                trace_get_payload_from_udp(udp, &rem);
                return size_t(slen-rem)+extra;
            }
            return SLEN_ERROR;
        };
    }
    return NULL;
}

static void usage(char *prog) {
    fprintf(stderr,"Usage:\n"
            "%s flags outputuri traceuri [traceuri...]\n"
            "  or\n"
            "%s flags -B baseuri traceuri [traceuri...]\n"
            "\n"
            "where flags are:\n"
            "[-B | --baseuri] baseuri\n"
            "    Output timestamped files to this baseuri\n"
            "[-F | --filter] bpf\n"
            "    Discard packets not matching the filter\n"
            "[-G | --rotate-seconds] seconds\n"
            "    Rotate output every so often, even if there are no packets\n"
            "[-S | --rotate-sizemb] sizeMB\n"
            "    Rotate output when it exceeds sizeMB\n"
            "[-H | --libtrace-help]\n"
            "    Print libtrace runtime documentation\n"
            "[-h | --help]\n"
            "    Print this help\n"
            "[-s | --snaplen] bytes\n"
            "    Capture this much of a packet (default=64). Examples:\n"
            "    -s 44                       sets the default snaplen to 44 bytes\n"
            "    -s 0 -s'1500:tcp'           sets the snaplen of tcp packets to 1500, or full capture otherwise\n"
            "    -s'UPTO_TCP:tcp port 53'    adjust snaplen not to go past tcp header for tcp port 53 packets\n"
            "[-U | --use-utc]\n"
            "    Use UTC in timestamping files\n"
            "[-v | --verbose]\n"
            "    Verbose output on stderr\n"
            "[-W | --watchfile] filename\n"
            "    Wait until the watchfile is created before proceeding with next segment\n"
            "[-w | --maxwait_ms] wait_ms\n"
            "    Spend at most this much time waiting on all interfaces per round-robin\n"
            "    Longer wait reduces disorders, but can cause drops.  This can be a float\n"
            "[-z | --compress-level] level\n"
            "    Sets compression level of output\n"
            "[-Z | --compress-type] type\n"
            "    Sets compression type\n"
            "[--file-ext=<extension>]\n"
            "    Sets output file extension (used with -B)\n"
            "[--relinquish-privileges=<username>]\n"
            "    Run capture as <username>\n"
            "[--pipeout=<command>]\n"
            "    Pipe output through <command> first\n"
            "[--rotate-period=<seconds>]\n"
            "    Rotate output on <second>'s boundary, e.g. for 3600 on hour boundary:\n"
            "    00:00:00, 01:00:00, 02:00:00, ...\n"
            "[--rotate-offset=<seconds>|rotate-phase=<seconds>]\n"
            "    Works with --rotate-period, to control output rotation offset, e.g for \n"
            "    --rotate-period=3600 --rotate-offset=1800 forces rotation at:\n"
            "    00:30:00, 01:30:00, 02:30:00, ...\n"

            , prog, prog);
    ::exit(1);
}

volatile int signalled = 0;

static void cleanup_signal(int sig) {
    if (sig == SIGCHLD) {
        int status = 0;
        waitpid(-1, &status, WNOHANG);
        return;
    }
    signalled = 1;
    trace_interrupt();
}


//struct time_order {
//    bool operator() (const libtrace_packet_t * lhs,
//                     const libtrace_packet_t * rhs) const {
//        return (trace_get_erf_timestamp(lhs) > trace_get_erf_timestamp(rhs));
//    };
//};


static const char * opt_seqnumfile = 0;
static const char * opt_pipe_arg[1024];

int
main(int argc, char *argv[]) {
    MTC_Input *input;
    int i=0;
    struct sigaction sigact;
    trace_option_compresstype_t compress_type = TRACE_OPTION_COMPRESSTYPE_NONE;
    list<snaplen_filter_t*> opt_snapfilters;
    string opt_val;
    int         opt_compress_level = -1;
    const char *opt_compress_type = NULL;
    int         opt_snaplen = 64;
    const char *opt_watchfile = NULL;
    const char *opt_relinquish = NULL;
    const char *opt_extension = NULL;
    const char *opt_filter = NULL;
    char       *opt_basename = NULL;
    char       *opt_pipeout = NULL;
    ulong       opt_segmentsize = 0;
    time_t      opt_rotatesec = 0;
    time_t      opt_rotatesec_period = 0;
    time_t      opt_rotatesec_phase = 0;
    double      opt_maxwait = MAXWAIT_MS;
    int         opt_verbose = MTC_Log::LOG_LEVEL_PANIC;
    bool        opt_useutc = false;

    //    uint64_t total_packets = 0;
    //priority_queue<struct libtrace_packet_t *,
    //                    vector<struct libtrace_packet_t *>,
    //                    time_order > pqueue;
    enum {
        OPT_RELINQUISH_PRIVS = 0x01f0,
        OPT_PIPEOUT,
        OPT_FILE_EXT,
        OPT_ROTATE_PERIOD,
        OPT_ROTATE_PHASE
    };

    MTC_Log tclog;
    tclog.set_log_level(opt_verbose);

    while (1) {
        int option_index;
        struct option long_options[] =
            {
                { "libtrace-help",  0, 0, 'H' },
                { "help",           0, 0, 'h' },
                { "baseuri",        1, 0, 'B' },
                { "rotate-seconds", 1, 0, 'G' },
                { "seqfile",        1, 0, 'N' },
                { "rotate-sizemb",  1, 0, 'S' },
                { "snaplen",        1, 0, 's' },
                { "use-utc",        0, 0, 'U' },
                { "verbose",        0, 0, 'v' },
                { "watchfile",      1, 0, 'W' },
                { "maxwait_ms",     1, 0, 'w' },
                { "compress-level", 1, 0, 'z' },
                { "compress-type",  1, 0, 'Z' },
                { "filter",         1, 0, 'F' },
                { "relinquish-privileges",
                  1, 0, OPT_RELINQUISH_PRIVS },
                { "pipeout",        1, 0, OPT_PIPEOUT },
                { "file-ext",       1, 0, OPT_FILE_EXT },
                { "rotate-period",  1, 0, OPT_ROTATE_PERIOD },
                { "rotate-offset",  1, 0, OPT_ROTATE_PHASE },
                { "rotate-phase",   1, 0, OPT_ROTATE_PHASE },
                { NULL,             0, 0, 0   },
            };

        int c = getopt_long(argc, argv, "HhF:B:G:N:S:s:Uvz:Z:W:w:",
                            long_options, &option_index);

        if (c == -1)
            break;

        switch (c) {
        case 'H':
            trace_help();
            ::exit(1);
            break;
        case 'h':
            usage(argv[0]);
            break;
        case 'B':
            opt_basename = optarg;
            break;
        case 'F':
            opt_filter = optarg;
            break;
        case 'G':
            opt_rotatesec = strtoul(optarg, NULL, 10);
            break;
        case 'N':
            opt_seqnumfile = optarg;
            break;
        case 'S':
            opt_segmentsize = strtoul(optarg, NULL, 10);
            break;
        case 's':
            opt_val = string(optarg);
            i = opt_val.find(":");
            if (size_t(i) != string::npos) {
                size_t extra = 0;
                slenfunc_t sfunc = NULL;
                string token  = opt_val.substr(0, i);
                string filter = opt_val.substr(i+1);
                char *ep;
                slen_t slen = slen_t(::strtoul(token.c_str(), &ep, 10));
                if (*ep != '\0') {
                    // token is not a number, it must be a mnemonic
                    string mnemonic = token;
                    i = token.find("+");
                    if (size_t(i) != string::npos) {
                        // mnemonic is followed by extra, as in "UPTO_TCP+128"
                        string extra_str = token.substr(i+1);
                        mnemonic = token.substr(0, i);
                        extra = size_t(::strtoul(extra_str.c_str(), &ep, 10));
                        if (extra_str[0] != '\0' && *ep != '\0') {
                            tclog.panic("Failed to parse snaplen mnemonic: '%s' is not a number, expected +<number>\n", extra_str.c_str());
                        }
                    }
                    sfunc = slen_name2func(mnemonic, extra);
                    if (sfunc == NULL) {
                        tclog.panic("Snaplen mnemonic %s is not supported; currently available are 'UPTO_IP', 'UPTO_IP6', 'UPTO_TCP', 'UPTO_UDP'\n", mnemonic.c_str());
                    }
                } else {
                    sfunc = slen_const(slen);
                }
                opt_snapfilters.insert(
                                       opt_snapfilters.end(),
                                       new snaplen_filter_t(trace_create_filter(filter.c_str()), sfunc)
                                       );
            } else {
                // default snaplen
                char *ep;
                opt_snaplen = ::strtoul(optarg, &ep, 10);
                if (*ep != '\0') {
                    tclog.panic("Failed to parse snaplen %s, not a number\n", optarg);
                }
            }
            break;
        case 'U':
            opt_useutc = true;
            break;
        case 'v':
            ++opt_verbose;
            tclog.set_log_level(opt_verbose);
            break;
        case 'W':
            opt_watchfile = optarg;
            break;
        case 'w':
            errno = 0;
            opt_maxwait = ::strtod(optarg, NULL);
            if (errno != 0) {
                tclog.panic("Cannot parse maxwait parameter.  Expected a float, got: %s\n", optarg);
            }
            break;
        case 'z':
            opt_compress_level = atoi(optarg);
            if (opt_compress_level<0 || opt_compress_level>9) {
                fprintf(stderr,"Compression level must be between 0 and 9\n");
                usage(argv[0]);
            }
            break;
        case 'Z':
            opt_compress_type = optarg;
            break;
        case OPT_RELINQUISH_PRIVS:
            opt_relinquish = optarg;
            break;
        case OPT_PIPEOUT:
            opt_pipeout = optarg;
            {
                int ii=0, cc=0;
                opt_pipe_arg[cc] = opt_pipeout;
                for (; opt_pipeout[ii] != '\0'; ++ii) {
                    if (opt_pipeout[ii] == ' ') {
                        opt_pipeout[ii] = '\0';
                        opt_pipe_arg[++cc] = &opt_pipeout[ii+1];
                    }
                }
                opt_pipe_arg[++c] = NULL;
            }
            break;
        case OPT_FILE_EXT:
            opt_extension = optarg;
            break;
        case OPT_ROTATE_PERIOD:
            opt_rotatesec_period = strtoul(optarg, NULL, 10);
            break;
        case OPT_ROTATE_PHASE:
            opt_rotatesec_phase = strtoul(optarg, NULL, 10);
            break;
        default:
            fprintf(stderr,"unknown option: %c\n",c);
            usage(argv[0]);
        }

    }

    if (opt_basename) {
        if (optind + 1 > argc)
            usage(argv[0]);
    } else {
        if (optind + 2 > argc)
            usage(argv[0]);
    }

    if (opt_compress_type == NULL && opt_compress_level >= 0) {
        fprintf(stderr, "Compression level set, but no compression type was defined, setting to gzip\n");
        compress_type = TRACE_OPTION_COMPRESSTYPE_ZLIB;
    } else if (opt_compress_type == NULL) {
        /* If a level or type is not specified, use the "none"
         * compression module */
        compress_type = TRACE_OPTION_COMPRESSTYPE_NONE;
    } else if (strncmp(opt_compress_type, "gz", 2) == 0 ||
               strncmp(opt_compress_type, "zlib", 4) == 0) {
        compress_type = TRACE_OPTION_COMPRESSTYPE_ZLIB;
    } else if (strncmp(opt_compress_type, "bz", 2) == 0) {
        compress_type = TRACE_OPTION_COMPRESSTYPE_BZ2;
    } else if (strncmp(opt_compress_type, "lzo", 3) == 0) {
        compress_type = TRACE_OPTION_COMPRESSTYPE_LZO;
    } else if (strncmp(opt_compress_type, "xz", 2) == 0) {
        compress_type = TRACE_OPTION_COMPRESSTYPE_LZMA;
    } else if (strncmp(opt_compress_type, "no", 2) == 0) {
        compress_type = TRACE_OPTION_COMPRESSTYPE_NONE;
    } else {
        tclog.panic("Unknown compression type: %s\n", opt_compress_type);
    }

    timeval now;
    ::gettimeofday(&now, NULL);

    MTC_Output *tco = 0;
    if (opt_basename) {
        // if basename is given, this is a uri
        // to a rotatable file and all our arguments are input uri's
        tco = new MTC_Output(0, opt_basename, now, tclog);
    } else {
        // if no basename, first argument is output uri
        tco = new MTC_Output(argv[optind++], 0, now, tclog);
    }
    //the rest are input uris


    sigact.sa_handler = cleanup_signal;
    sigemptyset(&sigact.sa_mask);
    //sigact.sa_flags = SA_RESTART;
    sigact.sa_flags = SA_RESTART;

    sigaction(SIGINT, &sigact, NULL);
    sigaction(SIGTERM,&sigact, NULL);
    sigaction(SIGCHLD, &sigact, NULL);



    int inputs = argc - optind;

    timeval maxwait_tv; //max wait per input per round, only if packets
    maxwait_tv.tv_sec = time_t(opt_maxwait / 1000. / inputs);
    maxwait_tv.tv_usec= time_t((opt_maxwait - maxwait_tv.tv_sec *1000.*inputs)*1000. / inputs);
    input = new MTC_Input[inputs];

    struct libtrace_filter_t *filter = NULL;
    if (opt_filter) {
        filter = trace_create_filter(opt_filter);
    }
    for (i = 0; i < inputs; ++i) {
        //libtrace_packet_t *p = trace_create_packet();
        const char *uri = argv[i+optind];
        libtrace_t *f = ::trace_create(uri);
        if (::trace_is_err(f)) {
            trace_perror(f, "trace_create");
            ::exit(1);
        }
        if (trace_set_event_realtime(f, true) < 0) {
            trace_get_err(f);
        }
        trace_set_snaplen(f, opt_snaplen);
        if (filter) {
            if (trace_config(f, TRACE_OPTION_FILTER, filter) != 0) {
                trace_perror(f, "Failed to setup filter for %s\n", uri);
                ::exit(1);
            }
        }
        input[i].in_ = f;
        input[i].uri_= uri;
        input[i].active_ = true;
        input[i].maxwait_ = maxwait_tv;
        if (trace_start(f) == -1) {
            trace_perror(f, "trace_start");
            ::exit(1);
        }
        libtrace_stat_t *stat = trace_get_statistics(input[i].in_, NULL);
        input[i].segment_drops_ = stat->dropped;
    }

    if (opt_relinquish) {
        struct passwd *pw = getpwnam(opt_relinquish);
        if (!pw) {
            tclog.panic("Can't find user '%s'\n", opt_relinquish);
        }
        if (initgroups(pw->pw_name, pw->pw_gid) != 0 ||
            setgid(pw->pw_gid) != 0 ||
            setuid(pw->pw_uid) != 0) {
            tclog.panic("Failed to drop root privileges: %s\n", strerror(errno));
        }
    }
    if (opt_seqnumfile) {
        tco->set_seqnumfile(opt_seqnumfile);
    }
    if (opt_compress_level >= 0 &&
        compress_type != TRACE_OPTION_COMPRESSTYPE_NONE) {
        tco->set_compression(compress_type, opt_compress_level);
    }

    if (opt_watchfile) {
        tco->set_watchfile(opt_watchfile);
    }
    if (opt_segmentsize) {
        tco->set_segmentsize(1024*1024*opt_segmentsize); //Mbytes to bytes
    }
    if (opt_rotatesec) {
        tco->set_rotatesec(opt_rotatesec);
    }
    if (opt_extension) {
        tco->set_extension(opt_extension);
    }

    tco->set_useutc(opt_useutc);
    if (opt_pipeout) {
        tco->set_pipeout((char* const*)opt_pipe_arg);
    }
    tco->set_inputs(input, inputs);


    fd_set rfds;

    //int maxfd = -1;
    FD_ZERO(&rfds);
    int active_inputs = inputs;
    libtrace_packet_t *p = 0;
    int ret=0;
    while (active_inputs > 0 && !signalled) {
        gettimeofday(&now, NULL);
        if ((opt_rotatesec && (now.tv_sec >= tco->last_rotated().tv_sec + opt_rotatesec)) ||
            (opt_rotatesec_period &&
             ((now.tv_sec - opt_rotatesec_phase)/opt_rotatesec_period >
              (tco->last_rotated().tv_sec - opt_rotatesec_phase)/opt_rotatesec_period))) {
            tco->rotate_trace(now); //force rotation by time
        }
        uint64_t ts;
        uint64_t mintime_erf = -1;
        int      mintime_idx = -1;
        int      sources = 0;
        for (i = 0; i < inputs; ++i) {
            if (!input[i].active_)
                continue;
            if (input[i].packet_) {
                ts = trace_get_erf_timestamp(input[i].packet_);
                ++sources;
                if (ts < mintime_erf) {
                    mintime_erf = ts;
                    mintime_idx = i;
                }
                continue; //already have a packet
            }
            if (p == 0)
                p = trace_create_packet();
            libtrace_eventobj_t evt = trace_event(input[i].in_, p);

            //uint64_t ts = trace_get_erf_timestamp(p);
            switch (evt.type) {
            case TRACE_EVENT_SLEEP:
                tclog.debug("sleep event on %s (%d, for %es)\n",
                            input[i].uri_, i, evt.seconds);
#if 0
                FD_ZERO(&rfds);
                FD_SET(evt.fd, &rfds);
                {
                    struct timeval tv;
                    tv.tv_sec = int(evt.seconds);
                    tv.tv_usec= (evt.seconds - tv.tv_sec)*1e6;
                    select(0, &rfds, NULL, NULL, &tv);
                }
#else
                //--i; //retry immediately
                continue;
#endif

            case TRACE_EVENT_IOWAIT:
                //fprintf(stderr, "iowait, %d\n", i);
                if ((input[i].maxwait_.tv_sec | input[i].maxwait_.tv_usec) == 0) {
                    //no more waiting!
                    continue;
                }
                FD_ZERO(&rfds);
                FD_SET(evt.fd, &rfds);
                ret = select(evt.fd + 1, &rfds, NULL, NULL, &input[i].maxwait_);
                if (ret == 0) {
                    continue; //timeout
                } else if (ret < 0) {
                    if (errno != EINTR)
                        tclog.warn("select returned error for wating on fd %d (%s)\n",
                                   evt.fd, input[i].uri_);
                    continue;
                }
                assert(ret == 1);
                evt = trace_event(input[i].in_, p);
                if (evt.type != TRACE_EVENT_PACKET) {
                    tclog.warn("event type: %d\n", evt.type);
                    //--i; //retry immediately
                    continue;
                }
                //assert(evt.type == TRACE_EVENT_PACKET);

                /* FALLTHROUGH */

            case TRACE_EVENT_PACKET:
                if (signalled) {
                    break;
                }
                {
                    uint16_t ethertype;
                    uint32_t remaining;
                    void *vp = trace_get_layer3(p, &ethertype, &remaining);
                    if (!vp || ethertype == 0xffff) {
                        tclog.warn("skipping non L3 (ethernet) packet on %s\n", input[i].uri_);
                        --i; /* xxx rerun the same input */
                        continue;
                    }
                }
                input[i].maxwait_ = maxwait_tv; //for next iteration, only if got a packet
                //fprintf(stderr, "pushed: %p\n", p);
                input[i].packet_ = p;
                ts = trace_get_erf_timestamp(p);
                ++sources;
                if (ts < mintime_erf) {
                    mintime_erf = ts;
                    mintime_idx = i;
                }
                //                ++total_packets;
                ++input[i].total_packets_;
                ++input[i].segment_packets_;
                if (input[i].prev_ts_ > ts) {
                    tclog.warn("disorder on input %d: %.6f, packet: %llu\n",
                               i,
                               (double)(input[i].prev_ts_-ts)/(1ULL<<32),
                               input[i].total_packets_);
                }
                input[i].prev_ts_ = ts;
                p = 0;
                continue;
            case TRACE_EVENT_TERMINATE:
                //end of trace
                tclog.debug("TERMINATE on %d (%s)\n", i, input[i].uri_);
                input[i].active_ = false;
                assert(input[i].packet_ == 0);
                --active_inputs;
                continue;
            default:
                fprintf(stderr, "Unknown event type occured\n");
                trace_perror(input[i].in_, "%s", argv[i+2]);
                ::exit(1);
            }

            //oldest_ts>trace_get_erf_timestamp(packet[i]))) {
        }
        if (mintime_idx == -1) {
            //fprintf(stderr, "no packets!\n");
            continue;
        }
        if (p) {
            trace_destroy_packet(p);
            p = 0;
        }
        //fprintf(stderr, "%d\n", sources);
        p = input[mintime_idx].packet_;

        //check if we need to rotate
        timeval ptv = trace_get_timeval(p);
        if ((opt_rotatesec && (ptv.tv_sec >= tco->last_rotated().tv_sec + opt_rotatesec)) ||
            (opt_rotatesec_period &&
             ((ptv.tv_sec - opt_rotatesec_phase)/opt_rotatesec_period >
              (tco->last_rotated().tv_sec - opt_rotatesec_phase)/opt_rotatesec_period))) {
            tco->rotate_trace(ptv); //force rotation by time
        }

        // adjust packet's snaplen if it's based on a filter
        for (list<snaplen_filter_t*>::iterator it = opt_snapfilters.begin();
             it != opt_snapfilters.end();
             ++it) {
            if (trace_apply_filter((*it)->filter, p)) {
                size_t slen = (*it)->sfunc(p);
                if (slen != SLEN_ERROR) {
                    trace_set_capture_length(p, (*it)->sfunc(p));
                } else {
                    tclog.debug("adaptive snaplen func return error, this usually means that the filter doesn't match the mnemonic\n");
                }
            }
        }
        tco->write_packet(p);

        input[mintime_idx].packet_ = 0;
    }
    if (p) {
        trace_destroy_packet(p);
        p = 0;
    }

    if (opt_verbose) {
        tco->dump_seg_stats();
        tco->dump_tot_stats();
    }
    //xxx make sure all packets are done
    gettimeofday(&now, NULL);
    tco->rotate_trace(now);

    for (i = 0; i < inputs; ++i) {
        libtrace_stat_t *stat = trace_get_statistics(input[i].in_, NULL);
        tclog.warn("closing input %d, total packets: %llu, drops: %lu\n",
                   i, input[i].total_packets_, stat->dropped);
        trace_destroy(input[i].in_);
        input[i].active_ = false;
        assert(signalled || (input[i].packet_ == 0));
    }

    // cleanup everything
    delete [] input;
    delete tco;
    for (list<snaplen_filter_t*>::iterator it = opt_snapfilters.begin();
         it != opt_snapfilters.end();
         ++it) {
        trace_destroy_filter((*it)->filter);
    }
    return 0;
}
